@@ -9,16 +9,6 @@
 from __future__ import print_function
 import os, sys, subprocess, tempfile
 
-# This module should not import elftools before setup_syspath() is called!
-# See the Hacking Guide in the documentation for more details.
-
-def setup_syspath():
-    """ Setup sys.path so that tests pick up local pyelftools before the
-        installed one when run from development directory.
-    """
-    if sys.path[0] != '.':
-        sys.path.insert(0, '.')
-
 
 def run_exe(exe_path, args=[], echo=False):
     """ Runs the given executable as a subprocess, given the
@@ -30,7 +20,7 @@ def run_exe(exe_path, args=[], echo=False):
         popen_cmd.insert(0, sys.executable)
     if echo:
       print('[cmd]', ' '.join(popen_cmd))
-    proc = subprocess.Popen(popen_cmd, stdout=subprocess.PIPE)
+    proc = subprocess.Popen(popen_cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
     proc_stdout = proc.communicate()[0]
     from elftools.common.py3compat import bytes2str
     return proc.returncode, bytes2str(proc_stdout)
@@ -42,16 +32,16 @@ def is_in_rootdir():
     return os.path.isdir('test') and os.path.isdir('elftools')
 
 
-def dump_output_to_temp_files(testlog, *args):
+def dump_output_to_temp_files(testlog, filename, option, *args):
     """ Dumps the output strings given in 'args' to temp files: one for each
-        arg.
+        arg. The filename and option arguments contribute to the file name,
+        so that one knows which test did the output dump come from.
     """
     for i, s in enumerate(args):
         fd, path = tempfile.mkstemp(
-                prefix='out' + str(i + 1) + '_',
+                prefix='out-%d-%s-%s-' % (i + 1, os.path.split(filename)[-1], option),
                 suffix='.stdout')
         file = os.fdopen(fd, 'w')
         file.write(s)
         file.close()
         testlog.info('@@ Output #%s dumped to file: %s' % (i + 1, path))
-
